import java.awt.*;
import java.io.*;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import amzi.ls.*;
import java.lang.*;
import java.util.*;

public class agentServer extends UnicastRemoteObject implements iAgent
{
    public logicEnd black;
    public boolean loggedIn=false;
    public String rules=null;
    public String delta=null;
    public agentFrame frame=null;
    public Vector others;
    public logicEnd my;
    public iBlackboard b;
    public File outputFile;
    public FileOutputStream outs;
    public PrintWriter outStream;
    public int nextLAInstance=0, ThreadGroupCount=0;

    public int login()
    {
        boolean standAlone = false; // se occorre far partire l'agente in modalit standalone
        if( black==null ) {
            printMessage( "Non sono ancora stati definiti i parametri della blackboard" );
            return -1;
        }
        System.out.println( my.name + " looking for blackboard " + black.name + " at " + black.address + ":" + black.port + "..." );
		try {
            b =( iBlackboard )Naming.lookup("rmi://"+black.address+":"+black.port+"/"+black.name);
        } catch (Exception e) {
            printMessage( "Impossibile trovare la blackboard! Assicurarsi che esista." );
            standAlone = true;
        }
        if( standAlone ) {
            System.out.println( my.name + " viene fatto partire in modalit standalone" );
            return -1;
        }
        System.out.println( my.name + " logging in blackboard " + black.name + " at " + black.address + ":" + black.port + "..." );
        loginThread t = new loginThread( this, b, my.name, my.address, my.port, delta );
        t.start();
        try {
            t.join();
        } catch ( InterruptedException e ) { System.out.println( "@@@ Errore nel join" ); }
        if( loggedIn = t.getResult() ) {
//            others = t.getVector(); // viene gi aggiornato da loginThread
            if( !updateReferences() ) System.out.println( "error binding agentServer" );
            return others.size();
        }
        return -1; // in caso di fallimento del login()
    }
    
    public boolean logout()
    {
        if( !loggedIn ) return true;
		try {
            b = ( iBlackboard )Naming.lookup( "rmi://" + black.address + ":" + black.port + "/" + black.name );
        } catch (Exception e) {
            e.printStackTrace();
            printMessage( "Impossibile trovare la blackboard! Assicurarsi che esista." );
            try { System.exit(1); } catch ( SecurityException se ) {
                se.printStackTrace();
                System.out.println( "@@@ Premere Ctrl+C per terminare l'esecuzione della JVM" );
                }
        }
        printMessage( my.name + " logging out from blackboard " + black.name + " at " + black.address + ":" + black.port + "..." );
        loginThread t = new loginThread( b, my.name );
        t.start();
        try {
            t.join();
        } catch ( InterruptedException e ) { System.out.println( "@@@ Errore nel join" ); }
        others = null;
        return !( loggedIn = !t.getResult() );
    }

    public Vector askRemoteCons( String goal, Vector abdIn ) throws RemoteException {
        Vector abdOut = new Vector();
        printMessage( "agentServer: ricevuta richiesta di consistenza per il termine '" + goal + "' con Lista_abducibili = ", abdIn );
        try {
            logicAgent lg = new logicAgent( this );
            // Esecuzione derivazione di consistenza
            abdOut = lg.askConsLogicAgent( goal, abdIn );
            lg.close();
        } catch ( LSException e ) { System.out.println( "@@@ " + e.GetMsg() ); }
        return abdOut;
    }

    public synchronized void updateAgentList( Vector bunch ) throws RemoteException {
        printMessage( "agentServer: updating agent list..." );
        boolean found=false;
        others.removeAllElements();
        for( int i=0; i<bunch.size(); i++ ) {
            if ( ( ( ( logicEnd )bunch.elementAt(i) ).name ).equals( my.name ) ) {
                found = true;
            }
            else others.addElement( ( logicEnd )bunch.elementAt(i) );
        }
        if( !found ) {
            printMessage( "agentServer: updateAgentList: IO non mi sono trovato nella lista" );
        }
        if( !updateReferences() ) System.out.println( "error binding agentServer" );
        printMessage( "...agent list updated." );
    }

    public void setReferenceFrame(agentFrame af)
    {
        frame = af;
    }
    public agentServer( String args[] ) throws RemoteException, IOException
    {
        super();
        System.out.println( "agentServer( String args[] )..." );
        // si attiva
        my=new logicEnd( args[0], Integer.parseInt( args[1] ) );
        System.out.println( "agentServer " + my.name + " activating on " + my.address + ":" + my.port + "..." );
        if( !my.bind( this ) ) {
            System.out.println( "binding dell'agente fallito" );
            System.exit(2);
        }
        // cerca la blackboard
        System.out.println( "agentServer( String args[] )..." );
        String bName, bAddress; int bPort;
        if( args.length>2 ) bPort = Integer.parseInt( args[2] ); else bPort=1099;
        if( args.length>4 ) bName = args[4]; else bName = "black";
        if( args.length>5 ) rules = args[5]; else rules = "default.xpl";
        if( args.length>6 ) delta = args[6]; else delta = new String( "[]" );
        if( args.length>3 ) {
            bAddress = args[3];
            black=new logicEnd( bName, bAddress, bPort );
        }
        else {
            black=new logicEnd( bName, bPort );
        }
        login();
        Date date = new Date();
        String outPathName = new String( "Output/out-" + date.getTime() + ".txt" );
        outputFile = new File( outPathName );
        outs = new FileOutputStream( outputFile );
        outStream = new PrintWriter( outs, true );
        System.out.println( date.toString() + ": creato l'agentServer " + my.name + " in ascolto su " + my.address + ":" + my.port );
        outStream.println( date.toString() + ": creato l'agentServer " + my.name + " in ascolto su " + my.address + ":" + my.port );
    }
    public static void main(String args[]) {
        System.out.println( "agentServer:main()..." );
        if( args.length<2 ) {
            System.out.println( "agentServer funziona a riga di comando:" );
            System.out.println( "occorre specificare il nome dell'agente e la porta su cui si mette in ascolto" );
            System.out.println( "java agentServer <name> <port> [<blackboard port> <blackboard address> <blackboard name>]" );
        }
        else try {
            /* Setting dei riferimenti: il server viene avviato tramite l'istanza
            af di AgentFrame una volta definiti tutti i parametri */
            agentServer as= new agentServer( args );
            agentFrame af = new agentFrame( as );
            af.setVisible( true );
            as.setReferenceFrame( af );
        } catch ( Exception e ) { e.printStackTrace(); }
    }

    public int nextInstance() {
        return nextLAInstance++;
    }

    public synchronized void println( String s ) {
        outStream.println( s );
    }

    public synchronized void println( String s, Vector v ) {
        if( v!=null ) outStream.println( s + v.toString() );
        else outStream.println( s + "[null]" );
    }

    public synchronized void printMessage( String m ) {
        if( frame!=null ) frame.printMessage( m );
        else System.out.println( m );
    }

    public synchronized void printMessage( String s, Vector v ) {
        if( frame!=null ) frame.printMessage( s, v );
        else {
            if( v!=null ) System.out.println( s + v.toString() );
            else System.out.println( s + "[null]" );
        }
    }

    public boolean setOthers( Vector others ) {
        this.others = others;
        return updateReferences();
    }

    private boolean updateReferences() {
        System.out.println( "loginThread: updateNaming..." );
        boolean ok=true;
        for( int i=0; i<others.size(); i++ ) {
            logicEnd le = ( logicEnd )others.elementAt(i);
            String serv = "rmi://" + le.address + ":" + le.port + "/" + le.name;
            try {
                le.agent = ( iAgent )Naming.lookup( serv );
            } catch ( RemoteException ignored ) {
                System.out.println( "@@@ RemoteException thrown within updateReferences at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
                System.out.println( "@@@ " + ignored.getMessage() );
                ok=false;
            } catch ( java.net.MalformedURLException ignored ) {
                System.out.println( "@@@ MalformedURLException thrown within updateReferences at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
                ok=false;
            } catch ( NotBoundException ignored ) {
                System.out.println( "@@@ NotBoundException thrown within updateReferences at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
                ok=false;
            }
        }
        System.out.println( "others: " + others.toString() );
        return ok;
    }   
}