import java.awt.*;
import java.io.*;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import java.util.Vector;
import java.util.Date;
import amzi.ls.*;
import java.lang.*;

public class blackboard extends java.rmi.server.UnicastRemoteObject implements iBlackboard
{
//    LogicServer ls;
    int policy=0; // 0 = In/Out; 1 = read/write
    String delta="[]";
    blackboardFrame frame=null;
    Vector agents=new Vector();
    logicEnd my;
    private boolean available=true; // si riferisce al delta
    private boolean availableForLogin=true; // si riferisce alla gestione del bunch
    // availableForLogin = true indica una situazione di stasi;
    // availableForLogin = false indica l'ingresso provvisorio di un agente nel bunch, 
    // il che blocca l'inizio di ogni risoluzione abduttiva.
    private File outputFile;
    private FileOutputStream outs;
    private PrintWriter outStream;

    public void exit() {
        try { System.exit(1); } catch ( SecurityException e ) {
            e.printStackTrace();
            System.out.println( "@@@ Premere Ctrl+C per chiudere la blackboard" );
            }
    }

    public blackboard( String args[] ) throws RemoteException, IOException
    {
        super();
        System.out.println( "blackboard( String args[] )..." );
        String bName; int bPort;
        if( args.length>0 ) bPort = Integer.parseInt( args[0] ); else bPort=1099;
        if( args.length>1 ) bName = args[1]; else bName = "black";
        my=new logicEnd( bName, bPort );
        my.bind( this );
        Date date = new Date();
        String outPathName = new String( "Output/out-" + date.getTime() + ".txt" );
        outputFile = new File( outPathName );
        outs = new FileOutputStream( outputFile );
        outStream = new PrintWriter( outs, true );
//        initLogicServer( ls );
        System.out.println( date.toString() + ": creata la blackboard " + my.name + " in ascolto su " + my.address + ":" + my.port );
        println( date.toString() + ": creata la blackboard " + my.name + " in ascolto su " + my.address + ":" + my.port );
    }

    public void setReferenceFrame( blackboardFrame bf )
    {
        frame = bf;
    }

    public synchronized String read() throws RemoteException {
        return read( false );
    }
    
    public synchronized boolean isAvailable() throws RemoteException {
        return available;
    }
    
    public synchronized boolean isAvailableForLogin() throws RemoteException {
        return availableForLogin;
    }

    public synchronized String read( boolean priority ) throws RemoteException {
        if( !priority ) {
            while( !availableForLogin ) {
                try {
                    wait();
                } catch ( InterruptedException e ) { System.out.println( "@@@ InterruptedException in blackboard.read()"); }
            }
            if( policy == 0 ) { // politica In/Out
                while ( available == false ) {
                    try {
                        wait();
                    } catch ( InterruptedException e ) { System.out.println( "@@@ InterruptedException in blackboard.read()"); }
                }
                available = false;
                printMessage( "Prelevato l'elemento " + delta + " dalla blackboard" );
            }
            else if( policy == 1 ) { // politica read/write
                printMessage( "Letto l'elemento " + delta + " dalla blackboard" );
            }
        }
        return delta;
    }

    public synchronized void unlock() throws RemoteException {
        available = true;
        notifyAll();
    }

    public synchronized void unlockLogin() throws RemoteException {
        availableForLogin = true;
        notifyAll();
    }

    /* vecchia versione (utilizzava il Prolog)

    public synchronized boolean write( String v ) throws RemoteException {
        if( policy==0 ) { // politica In/Out
            delta = v;
            available = true;
            frame.messages.append( "Inserito l'elemento " + delta + " nella blackboard\n" );
            notifyAll();
            return true;
        }
        else { // politica read/write
            long t=0;
            try {
                t = ls.ExecStr( "aggiorna_write(" + delta + "," + v + ",New)" );
            } catch (LSException e) { System.out.println( e.GetMsg() ); }
            if ( t>0 )
            {
            try {
                    long risultato = ls.GetArg( t, 3 ); //risultato contiene la lista degli abducibili trovati
                    delta = ls.TermToStr( risultato, 10000 );
                } catch ( LSException e ) { System.out.println( e.GetMsg() ); }
                return true;
            }
            else return false;
        }
    }
    */

    // nuova versione della write:
    public synchronized boolean write( String v ) throws RemoteException
    {
        boolean result = false;
        if( policy==1 )  // politica read/write
        {
            result = h.compatible( delta, v );
            if( result ) delta = h.union( delta, v );
        }
        else if( policy==0 )  // politica In/Out
        {
            delta = v;
            available = true;
            notifyAll();
            result = true;
        }
        if( result ) frame.messages.append( "Inserito l'elemento " + delta + " nella blackboard\n" );
        return result;
    }

    public String getName() {
        String n = new String( my.name );
        return n;
    }

    public int getPolicy() {
        // politica di scrittura (0 = In/Out, 1 = read/write)
        return policy;
    }

    // la getIn viene utilizzata dopo la requestLogin, cio dopo aver effettuato tutti i controlli.
    // si usa getIn invece di tempLogin se non  necessario fare dei controlli su delta.
    public synchronized Vector getIn( String name, String address, int port ) {
        System.out.println( "getIn(" + name + ")..." );
        Vector result = new Vector( agents.size() );
        for( int i=0; i<agents.size(); i++ )
            result.addElement( ( logicEnd )agents.elementAt(i) );
        logicEnd al = new logicEnd( name, address, port );
        System.out.println( al.toString() );
        agents.addElement( ( logicEnd )al );
        for( int i=0; i<agents.size(); i++ ) {
            logicEnd li = ( logicEnd )agents.elementAt(i);
            if ( !li.name.equals( name ) ) {
                // devo notificare la nuova lista agli altri agenti
                String serv="rmi://"+li.address+":"+li.port+"/"+li.name;
                try {
                    System.out.println( "Notifico la nuova lista a " + li.name + "@" + li.address + ":" + li.port );
                    iAgent a = ( iAgent )Naming.lookup( serv );
                    a.updateAgentList( agents );
                } catch ( Exception e ) { e.printStackTrace(); }
            }
        }
        printMessage( "L'agente " + name + " e` stato introdotto nel bunch..." );
        printMessage( agents.toString() );
        availableForLogin = true;
        notifyAll();
        return result;
    }

    // la tempLogin viene utilizzata dopo la requestLogin, cio dopo aver effettuato tutti i controlli.
    // so gi che i parametri sono corretti, non c' un omonimo, e non c' un altro agente che
    // ha fatto la stessa richiesta.
    // risultato: la lista degli altri agenti (che va messa in 'others')
    public synchronized Vector tempLogin( String name, String address, int port ) {
        System.out.println( "tempLogin(" + name + ")..." );
        Vector result = new Vector( agents.size() ); // lista corrente da restituire all'agente entrante
        for( int i=0; i<agents.size(); i++ ) {
            // in requestLogin() ho gi controllato che non ci sia un agente con lo stesso nome
            result.addElement( ( logicEnd )agents.elementAt(i) );
        }
        logicEnd al = new logicEnd( name, address, port );
        System.out.println( al.toString() );
        agents.addElement( ( logicEnd )al );
        for( int i=0; i<agents.size(); i++ ) {
            logicEnd li = ( logicEnd )agents.elementAt(i);
            if ( !li.name.equals( name ) ) {
                // devo notificare la nuova lista agli altri agenti
                String serv="rmi://"+li.address+":"+li.port+"/"+li.name;
                try {
                    System.out.println( "Notifico la nuova lista a " + li.name + "@" + li.address + ":" + li.port );
                    iAgent a = ( iAgent )Naming.lookup( serv );
                    a.updateAgentList( agents );
                } catch ( Exception e ) { e.printStackTrace(); }
            }
        }
        printMessage( "L'agente " + name + " e` stato introdotto (temporaneamente) nel bunch..." );
        System.out.println( agents.toString() );
        return result;
    }

    // risultato: null se l'agente non pu entrare
    // altrimenti, delta.
    public synchronized String requestLogin( String name, String address, int port ) {
        // mi metto in attesa della blackboard per il login
        while( !availableForLogin ) {
            try {
                wait();
            } catch ( InterruptedException e ) { System.out.println( "@@@ InterruptedException in blackboard.getIn()"); }
        }
        availableForLogin = false;
        System.out.println( "Controllo la correttezza dei parametri di login" );
        if( name==null || address==null || port<1024 ) {
            System.out.println( "Name, address o port e` specificata male. Impossibile procedere al login." );
            return null;
        }
        System.out.println( "requestLogin(" + name + ")..." );
        boolean found=false;
        for( int i=0; i<agents.size() && !found; i++ ) {
            if ( ( ( ( logicEnd )agents.elementAt(i) ).name ).equals( name ) ) {
                found = true;
            }
        }
        if( found ) {
            System.out.println( "E` gia` presente un agente di nome " + name + ": non  possibile procedere oltre al login dell'agente" );
            availableForLogin = true;
            notifyAll();
            // a questo punto si potrebbe verificare che sia vivo, e in
            // caso contrario sostituire il vecchio con il nuovo
            return null;
        }
        System.out.println( "Ora dovrei aspettare che le computazioni finiscano" );
        return delta;
    }

    public synchronized void confirmLogin( String name ) throws RemoteException {
        printMessage( "L'agente " + name + " e` stato confermato nel bunch." );
        availableForLogin = true;
        notifyAll();
    }
    
    public synchronized void rejectLogin( String name ) throws RemoteException {
        printMessage( "L'agente " + name + " non e` stato confermato nel bunch." );
        availableForLogin = true;
        notifyAll();
        getOut( name );
    }
    
    public synchronized boolean getOut( String name ) {
        boolean found=false;
        for( int i=0; i<agents.size() && !found; i++ ) {
            if ( ( ( ( logicEnd )agents.elementAt(i) ).name ).equals( name ) ) {
                found = true;
                agents.removeElementAt(i);
            }
        }
        if( found ) printMessage( "L'agente " + name + " e` stato tolto dal bunch..." );
        else printMessage( "L'agente " + name + " non e` stato trovato nel bunch..." );
        for( int i=0; i<agents.size(); i++ ) {
            logicEnd li = ( logicEnd )agents.elementAt(i);
            if ( !li.name.equals( name ) ) {
                // devo notificare la nuova lista agli altri agenti
                String serv="rmi://"+li.address+":"+li.port+"/"+li.name;
                try {
                    System.out.println( "Notifico la nuova lista a " + li.name + "@" + li.address + ":" + li.port );
                    iAgent a = ( iAgent )Naming.lookup( serv );
                    a.updateAgentList( agents );
                } catch ( Exception e ) { e.printStackTrace(); }
            }
        }
        System.out.println( agents.toString() );
        return found;
    }

    public static void main( String args[] ) {
        System.out.println( "blackboard:main()..." );
        try {
            /* Setting dei riferimenti: il server viene avviato tramite l'istanza
            af di AgentFrame una volta definiti tutti i parametri */
            blackboard b= new blackboard( args );
            blackboardFrame bf = new blackboardFrame( b );
            bf.setVisible(true);
            b.setReferenceFrame(bf);
        } catch ( Exception e ) { e.printStackTrace(); }
    }

    private void initLogicServer( LogicServer ls )
    {
        try {
            ls = new LogicServer();
            ls.Init( "" );
            ls.AddPred( "prompt", 1, "blackboard", "prompt", this );
            ls.Load( "compatibile" );

            // test del logic server:
            String parameters = "[a,not b]" + "," + "[a,not b]" + ",New", tempDelta="start value";
            long t = ls.ExecStr( "aggiorna_write(" + parameters + ")" );
            if ( t>0 ) // scrivo il delta
            {
                try {
                    long risultato = ls.GetArg( t, 3 ); //risultato contiene la lista degli abducibili trovati
                    tempDelta = ls.TermToStr( risultato, 10000 );
                } catch ( LSException e ) { System.out.println( e.GetMsg() ); }
                System.out.println( "aggiorna_write(" + parameters + ") ha dato risultato " + tempDelta );
            }
            else
            {
                System.out.println( "aggiorna_write(" + parameters + ") ha dato risultato negativo." );
            }
            // fine del test

        } catch ( LSException e ) { System.out.println( e.GetMsg() ); }
    }

    public synchronized void println( String s ) {
        outStream.println( s );
    }

    public synchronized void printMessage( String m ) {
        frame.printMessage( m );
    }

    public synchronized void printMessage( String m, Vector v ) {
        frame.printMessage( m, v );
    }
}