/* Metainterprete per la realizzazione della derivazione abduttiva
e della derivazione di consistenza*/

/* Abductive derivation */

start_abductive_derivation(Goal,AbdIn,AbdOut):-
     prompt('cominciata la derivazione abduttiva'),
     solve_abd(Goal,AbdIn,AbdOut).

/* normal resolution step */

solve_abd([],Abd,Abd):-!.

solve_abd([A|B],AbdIn,AbdOut):-!,
     solve_abd(A,AbdIn,Abd1),
     solve_abd(B,Abd1,AbdOut).

solve_abd(A,AbdIn,AbdOut):-
     rule(A,Body), /* Nota: in questo modo, il fatto che A sia gi inconsistente con 
                      AbdIn pu non venire rilevato, nel caso in cui esista una regola
                      che abbia A in testa. Questo funziona bene se forziamo le KB a
                      non avere abducibili in testa */
     solve_abd(Body,AbdIn,AbdOut).
/* end of normal resolution steps */

/* (1) A has already been abduced */
solve_abd(A,Abd,Abd):-
     abducible(A),
     member(A,Abd),!.

/* (2) A is abducibles but neither A nor not(A) have already been abduced */
solve_abd(not(A),AbdIn,AbdOut):-
     abducible(not(A)),!,
     not(member(A,AbdIn)),
     solve_con(not(A),[not(A)|AbdIn],LocalAbdOut),
          /* prima possibilita' di fallimento: un agente puo' violare i propri vincoli 
             quindi ritornera la lista vuota e se ne accorge la procedura Java 
             che realizza ask_remote_cons/3 */
     ask_remote_cons(not(A),LocalAbdOut,AbdOut1),
          /* seconda possibilita' di fallimento: esiste un conflitto nell'unione delle
             liste di ritorno che i vari agenti hanno fornito */ 
     no_conflicts(AbdOut1,AbdOut),!.

test_write_blackboard(New):-
     prompt('test_write_blackboard: scrivo il risultato sulla blackboard'),
     aggiorna(New).  

solve_abd(A,AbdIn,AbdOut):-
     abducible(A),!,
     not(member(not(A),AbdIn)),
     solve_con(A,[A|AbdIn],LocalAbdOut),
     ask_remote_cons(A,LocalAbdOut,AbdOut1),  
     no_conflicts(AbdOut1,AbdOut),!.

no_conflicts(ListIn,ListOut):-comp(ListIn,[],ListOut).

/*
no_conflicts(List,[]):-
     prompt('Trovato un conflitto del tipo [..,A,not(A),..] nella lista risultato'),
     prompt(List),
     fail.
*/

comp([],L,L).
comp([H|Tail],ListApp,ListOut):-
     not(member(not(H),Tail)),
     not(member(H,ListApp)),
     comp(Tail,[H|ListApp],ListOut).
comp([H|Tail],ListApp,ListOut):-
     not(member(not(H),Tail)),
     comp(Tail,ListApp,ListOut).
     
abducible(not(A)):-
     !,abducibles(AbdSet),
     (member(not(A),AbdSet) ; member(A,AbdSet) ).

abducible(A):-
     abducibles(AbdSet),
     (member(A,AbdSet) ; member(not(A),AbdSet) ).                        



/* Consistency derivation */

solve_con(Goal,AbdIn,AbdOut):-
     findall(Ic,get_ic(Goal,Ic),List_of_ic),!,
     add_ic(Goal,List_of_ic,All_ic),
     solve_con2(All_ic,AbdIn,AbdOut).

/* if there are no ics, then Goal can be abducted */
solve_con(_,Abd,Abd):-!.

get_ic(Goal,Ic):-
     ic(X),member(Goal,X),delete(X,Goal,Ic).

delete([],_,[]):-!.
delete([H|Rest],H,Rest):-!.
delete([H|Rest],Goal,[H|Rest1]):-delete(Rest,Goal,Rest1).

add_ic(not(Goal),List,[[Goal]|List]):-!.
add_ic(Goal,List,[[not(Goal)]|List]).

/* all the ic are satisfied */
solve_con2([],Abd,Abd):-!.

/* one of the ic has been violated */
solve_con2([[]|_],_,_):-
     !,
    % prompt('Un vincolo di consistenza e'' stato violato'),
     fail.

/* (3) first try to use resolution */
solve_con2([[L1|Rest_l]|Rest_ic],AbdIn,AbdOut):-
     setof(Body,rule(L1,Body),List_of_Bodies),!,
     append_all(List_of_Bodies,Rest_l,New_list),
     append(New_list,Rest_ic,New_ic),
     solve_con2(New_ic,AbdIn,AbdOut).

/* (4) L1 has already been abduced */
solve_con2([[L1|Rest_l]|Rest_ic],AbdIn,AbdOut):-
     member(L1,AbdIn),!,
     solve_con2([Rest_l|Rest_ic],AbdIn,AbdOut).

/* (5) not(L1) has been abduced */
solve_con2([[not(L1)|_]|Rest_ic],AbdIn,AbdOut):-
     member(L1,AbdIn),!,
     (Rest_ic\==[]->solve_con2(Rest_ic,AbdIn,AbdOut);
      AbdIn=AbdOut).

solve_con2([[L1|_]|Rest_ic],AbdIn,AbdOut):-
     member(not(L1),AbdIn),!,
     (Rest_ic\==[]->solve_con2(Rest_ic,AbdIn,AbdOut);
      AbdIn=AbdOut).

/* (6) L1 is abducible but neither L1 nor not(L1) have already been abduced */
solve_con2([[not(L1)|_]|Rest_ic],AbdIn,AbdOut):-
     abducible(not(L1)),!,
     solve_abd(L1,AbdIn,AbdOut1),
     (Rest_ic\==[]->solve_con2(Rest_ic,AbdOut1,AbdOut);
      AbdOut=AbdOut1).

solve_con2([[L1|_]|Rest_ic],AbdIn,AbdOut):-
     abducible(L1),!,
     solve_abd(not(L1),AbdIn,AbdOut1),
     (Rest_ic\==[]->solve_con2(Rest_ic,AbdOut1,AbdOut);
      AbdOut=AbdOut1).    

/* (7) L1 is not abducible and has no successful rule */
solve_con2([[_|_]|Rest_ic],AbdIn,AbdOut):-
     (Rest_ic\==[]->solve_con2(Rest_ic,AbdIn,AbdOut);
      AbdIn=AbdOut).    

append_all([],_,[]).

append_all([Body|Rest_b],Rest_l,[New_clause|Rest_c]):-
     append(Body,Rest_l,New_clause),
     append_all(Rest_b,Rest_l,Rest_c).

append([],X,X).
append([A|X],Y,[A|Z]):-append(X,Y,Z).


member(X,[X|_]):-!.
member(X,[_|Z]):-member(X,Z).


/*        -------------
     risoluzione abduttiva locale 
          -------------
*/

set_of_local_abductive_derivation(Goal,AbdIn,ListAbdOut):-
     setof(AbdOut,start_local_abductive_derivation(Goal,AbdIn,AbdOut),ListAbdOut).

start_local_abductive_derivation(Goal,AbdIn,AbdOut):-
     prompt(['cominciata la derivazione abduttiva locale',Goal,AbdIn,AbdOut]),
     solve_local_abd(Goal,AbdIn,AbdOut).
%     prompt(['terminata con successo la derivazione abduttiva locale',Goal,AbdIn,AbdOut]).

/* normal resolution step */

% solve_local_abd-0-I
solve_local_abd([],Abd,Abd):-
%     prompt(['try solve_local_abd-0-I',[],Abd,Abd]),
     !.

% solve_local_abd-0-II
solve_local_abd([A|B],AbdIn,AbdOut):-!,
%     prompt(['try solve_local_abd-0-II',[A|B],AbdIn,AbdOut]),
     solve_local_abd(A,AbdIn,Abd1),
     solve_local_abd(B,Abd1,AbdOut).
%     prompt(['end solve_local_abd-0-II',[A|B],AbdIn,AbdOut]).

% solve_local_abd-0-III
solve_local_abd(A,AbdIn,AbdOut):-
%     prompt(['try solve_local_abd-0-III',A,AbdIn,AbdOut]),
     rule(A,Body), /* Nota: in questo modo, il fatto che A sia gi inconsistente con 
                      AbdIn pu non venire rilevato, nel caso in cui esista una regola
                      che abbia A in testa. Questo funziona bene se forziamo le KB a
                      non avere abducibili in testa */
     solve_local_abd(Body,AbdIn,AbdOut).
%     prompt(['end solve_local_abd-0-III',A,AbdIn,AbdOut]).
/* end of normal resolution steps */

/* (1) A has already been abduced */
solve_local_abd(A,Abd,Abd):-
%     prompt(['try solve_local_abd-1',A,Abd,Abd]),
     abducible(A),
     member(A,Abd),!.
%     prompt(['end solve_local_abd-1',A,Abd,Abd]).

/* (2) A is abducibles but neither A nor not(A) have already been abduced */
solve_local_abd(not(A),AbdIn,AbdOut):-
%     prompt(['try solve_local_abd-2-I',not(A),AbdIn,AbdOut]),
     abducible(not(A)),!,
     not(member(A,AbdIn)),
     solve_local_con(not(A),[not(A)|AbdIn],AbdOut),!.
%     prompt(['end solve_local_abd-2-I',not(A),AbdIn,AbdOut]).

solve_local_abd(A,AbdIn,AbdOut):-
%     prompt(['try solve_local_abd-2-II',A,AbdIn,AbdOut]),
     abducible(A),!,
     not(member(not(A),AbdIn)),
     solve_local_con(A,[A|AbdIn],AbdOut),!.
%     prompt(['end solve_local_abd-2-II',A,AbdIn,AbdOut]).


/* Consistency derivation */

solve_local_con(Goal,AbdIn,AbdOut):-
%     prompt(['try solve_local_con-0-I',Goal,AbdIn,AbdOut]),
     findall(Ic,get_ic(Goal,Ic),List_of_ic),!,
     add_ic(Goal,List_of_ic,All_ic),
     solve_local_con2(All_ic,AbdIn,AbdOut).
%     prompt(['end solve_local_con-0-I',Goal,AbdIn,AbdOut]).

/* if there are no ics, then Goal can be abducted */
solve_local_con(_,Abd,Abd):-
%     prompt(['try solve_local_con-0-II',_,Abd,Abd]),
     !.

/* all the ic are satisfied */
solve_local_con2([],Abd,Abd):-
%     prompt(['try solve_local_con2-0-I',[],Abd,Abd]),
     !.

/* one of the ic has been violated */
solve_local_con2([[]|_],_,_):-!,
%     prompt(['try solve_local_con2-0-II',[[]|_],_,_]),
%     prompt('Un vincolo di consistenza e'' stato violato'),
     fail.

/* (3) first try to use resolution */
solve_local_con2([[L1|Rest_l]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-3',[[L1|Rest_l]|Rest_ic],AbdIn,AbdOut]),
     setof(Body,rule(L1,Body),List_of_Bodies),!,
     append_all(List_of_Bodies,Rest_l,New_list),
     append(New_list,Rest_ic,New_ic),
     solve_local_con2(New_ic,AbdIn,AbdOut).
%     prompt(['end solve_local_con2-3',[[L1|Rest_l]|Rest_ic],AbdIn,AbdOut]).

/* (4) L1 has already been abduced */
solve_local_con2([[L1|Rest_l]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-4',[[L1|Rest_l]|Rest_ic],AbdIn,AbdOut]),
     member(L1,AbdIn),!,
     solve_local_con2([Rest_l|Rest_ic],AbdIn,AbdOut).
%     prompt(['end solve_local_con2-4',[[L1|Rest_l]|Rest_ic],AbdIn,AbdOut]).

/* (5) not(L1) has been abduced */
solve_local_con2([[not(L1)|_]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-5-I',[[not(L1)|_]|Rest_ic],AbdIn,AbdOut]),
     member(L1,AbdIn),!,
     (Rest_ic\==[]->solve_local_con2(Rest_ic,AbdIn,AbdOut);
      AbdIn=AbdOut).
%     prompt(['end solve_local_con2-5-I',[[not(L1)|_]|Rest_ic],AbdIn,AbdOut]).

solve_local_con2([[L1|_]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-5-II',[[L1|_]|Rest_ic],AbdIn,AbdOut]),
     member(not(L1),AbdIn),!,
     (Rest_ic\==[]->solve_local_con2(Rest_ic,AbdIn,AbdOut);
      AbdIn=AbdOut).
%     prompt(['end solve_local_con2-5-II',[[L1|_]|Rest_ic],AbdIn,AbdOut]).

/* (6) L1 is abducible but neither L1 nor not(L1) have already been abduced */
solve_local_con2([[not(L1)|_]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-6-I',[[not(L1)|_]|Rest_ic],AbdIn,AbdOut]),
     abducible(not(L1)),!,
     solve_local_abd(L1,AbdIn,AbdOut1),
     (Rest_ic\==[]->solve_local_con2(Rest_ic,AbdOut1,AbdOut);
      AbdOut=AbdOut1).
%     prompt(['end solve_local_con2-6-I',[[not(L1)|_]|Rest_ic],AbdIn,AbdOut]).

solve_local_con2([[L1|_]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-6-II',[[L1|_]|Rest_ic],AbdIn,AbdOut]),
     abducible(L1),!,
     solve_local_abd(not(L1),AbdIn,AbdOut1),
     (Rest_ic\==[]->solve_local_con2(Rest_ic,AbdOut1,AbdOut);
      AbdOut=AbdOut1).
%     prompt(['end solve_local_con2-6-II',[[L1|_]|Rest_ic],AbdIn,AbdOut]).

/* (7) L1 is not abducible and has no successful rule */
solve_local_con2([[_|_]|Rest_ic],AbdIn,AbdOut):-
%     prompt(['try solve_local_con2-7',[[_|_]|Rest_ic],AbdIn,AbdOut]),
     (Rest_ic\==[]->solve_local_con2(Rest_ic,AbdIn,AbdOut);
      AbdIn=AbdOut).
%     prompt(['end solve_local_con2-7',[[_|_]|Rest_ic],AbdIn,AbdOut]).
